"""
This module holds the constants used in the other modules.

"""

CACHE_PATH = "cache"
TS_ORDINAL = "ordinal"
TS_TIMESTAMP = "timestamp"

FREQ_D = "d"
FREQ_W = "w"
FREQ_M = "m"
FREQ_Q = "q"
FREQ_Y = "y"

FREQ_H = "h"
FREQ_MIN = "min"
FREQ_SEC = "sec"

FREQ_DAYTYPES = (FREQ_D, FREQ_W, FREQ_M, FREQ_Q, FREQ_Y)
FREQ_IDAYTYPES = (FREQ_H, FREQ_MIN, FREQ_SEC)
